import sqlite3
import pytest
from pluggy import PluginManager
from src.chercher.db import init_db
from src.chercher import hookspecs


@pytest.fixture
def plugin_manager():
    pm = PluginManager("chercher")
    pm.add_hookspecs(hookspecs)

    return pm


@pytest.fixture
def test_db(tmp_path):
    db_path = tmp_path / "chercher_test.db"
    conn = sqlite3.connect(db_path)
    init_db(conn)

    yield conn
    conn.close()



from typing import Generator
import pluggy
from chercher.models import Document

hookspec = pluggy.HookspecMarker("chercher")
hookimpl = pluggy.HookimplMarker("chercher")


@hookspec
def ingest(uri: str, settings: dict) -> Generator[Document, None, None]:
    raise NotImplementedError



import sys
import sqlite3
import textwrap
import click
from loguru import logger
import pluggy
from rich.console import Console
from rich.table import Table
from chercher.plugin_manager import get_plugin_manager
from chercher.settings import Settings, APP_NAME, APP_DIR
from chercher.db import init_db, db_connection

console = Console()

logger.remove()
logger.add(
    sys.stderr,
    colorize=True,
    format="<green>{time:HH:mm:ss}</green> | <level>{level}</level> | {message}",
    level="INFO",
)
logger.add(
    APP_DIR / "chercher_errors.log",
    rotation="10 MB",
    retention="15 days",
    level="ERROR",
)

settings = Settings()


@click.group(help=settings.description)
@click.version_option(
    version=settings.version,
    message="v%(version)s",
    package_name=APP_NAME,
)
@click.pass_context
def cli(ctx: click.Context) -> None:
    with db_connection(settings.db_url) as conn:
        logger.debug("initializing the database")
        init_db(conn)

    ctx.ensure_object(dict)
    ctx.obj["settings"] = settings
    ctx.obj["db_url"] = settings.db_url
    ctx.obj["pm"] = get_plugin_manager()


def _index(conn: sqlite3.Connection, uris: list[str], pm: pluggy.PluginManager) -> None:
    cursor = conn.cursor()

    for uri in uris:
        try:
            for documents in pm.hook.ingest(uri=uri, settings=dict(settings)):
                for doc in documents:
                    try:
                        cursor.execute(
                            """
                    INSERT INTO documents (uri, body, metadata) VALUES (?, ?, ?)
                    """,
                            (doc.uri, doc.body, "{}"),
                        )
                        conn.commit()
                        logger.info(f'document "{uri}" indexed')
                    except sqlite3.IntegrityError:
                        logger.warning(f'document "{uri}" already exists')
                    except Exception as e:
                        logger.error(
                            f"something went wrong while indexing '{uri}': {e}"
                        )
        except Exception as e:
            logger.error(f"something went wrong while trying to index documents: {e}")


@cli.command(help="Index documents, webpages and more.")
@click.argument("uris", nargs=-1)
@click.pass_context
def index(ctx: click.Context, uris: list[str]) -> None:
    pm = ctx.obj["pm"]
    db_url = ctx.obj["db_url"]

    if not pm.list_name_plugin():
        logger.warning("No plugins registered!")
        return

    with db_connection(db_url) as conn:
        _index(conn, uris, pm)


@cli.command(help="Seach for documents matching your query.")
@click.argument("query")
@click.option(
    "-l",
    "--limit",
    type=int,
    default=5,
    help="Number of results.",
)
@click.pass_context
def search(ctx: click.Context, query: str, limit: int) -> None:
    db_url = ctx.obj["db_url"]

    with db_connection(db_url) as conn:
        cursor = conn.cursor()

        sql_query = """
            SELECT uri, substr(body, 0, 300)
            FROM documents
            WHERE ROWID IN (
                SELECT ROWID
                FROM documents_fts
                WHERE documents_fts MATCH ?
                ORDER BY bm25(documents_fts)
                LIMIT ?
            )
            """

        cursor.execute(sql_query, (query, limit))
        results = cursor.fetchall()

        if not results:
            console.print(f"No results found for: '{query}'")
            return

        for result in results:
            console.print(f"[link={result[0]}]{result[0]}[/]")
            console.print(
                f"{textwrap.shorten(result[1], width=280, placeholder='...')}\n",
                highlight=False,
            )


@cli.command(help="List out all the registered plugins and their hooks.")
@click.pass_context
def plugins(ctx: click.Context) -> None:
    pm = ctx.obj["pm"]
    plugins = dict(pm.list_plugin_distinfo())

    table = Table(title="plugins")
    table.add_column("name")
    table.add_column("hooks")

    for plugin in plugins:
        hooks = [h.name for h in pm.get_hookcallers(plugin)]
        hooks_str = ", ".join(hooks)
        table.add_row(plugin.__name__, hooks_str)

    console.print(table)



from typing import Generator
from faker import Faker
from src.chercher.cli import _index
from src.chercher import hookimpl, Document

fake = Faker()


class BadPlugin:
    @hookimpl
    def ingest(self, uri: str) -> Generator[Document, None, None]:
        if not uri.endswith(".txt"):
            raise Exception("Ooops!")


class DummyTxtPlugin:
    @hookimpl
    def ingest(self, uri: str) -> Generator[Document, None, None]:
        if uri.endswith(".txt"):
            yield Document(uri=uri, body="", metadata={})


def test_index_valid_txt(test_db, plugin_manager):
    plugin_manager.register(DummyTxtPlugin())
    uri = fake.file_path(depth=3, extension="txt")
    _index(test_db, [uri], plugin_manager)

    cursor = test_db.cursor()
    cursor.execute("SELECT * FROM documents")
    assert cursor.fetchone() is not None


def test_index_with_exception(test_db, plugin_manager):
    plugin_manager.register(DummyTxtPlugin())
    plugin_manager.register(BadPlugin())
    uris = [
        fake.file_path(depth=3, extension="pdf"),
        fake.file_path(depth=3, extension="txt"),
        fake.file_path(depth=3, extension="epub"),
    ]

    _index(test_db, uris, plugin_manager)

    cursor = test_db.cursor()
    cursor.execute("SELECT * FROM documents")
    documents = cursor.fetchall()

    assert len(documents) == 1


def test_index_same_document_multiple_times(test_db, plugin_manager):
    plugin_manager.register(DummyTxtPlugin())
    uri = fake.file_path(depth=3, extension="txt")

    _index(test_db, [uri, uri], plugin_manager)

    cursor = test_db.cursor()
    cursor.execute("SELECT * FROM documents WHERE uri = ?", (uri,))
    documents = cursor.fetchall()

    assert len(documents) == 1
    assert documents[0][0] == uri



from typing import Generator
from faker import Faker
from src.chercher import hookimpl, Document

fake = Faker()


class DummyPlugin:
    @hookimpl
    def ingest(self, uri: str) -> Generator[Document, None, None]:
        yield Document(uri=uri, body="", metadata={})


def test_dummy_ingest_plugin(plugin_manager):
    plugin_manager.register(DummyPlugin())
    uri = fake.file_path(depth=3)

    for documents in plugin_manager.hook.ingest(uri=uri):
        for doc in documents:
            assert doc.uri == uri



from importlib.metadata import version
from typing import Type, Tuple
from pathlib import Path
from click import get_app_dir
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
)

APP_NAME = "chercher"
APP_DIR = Path(get_app_dir(app_name=APP_NAME))
CONFIG_FILE_PATH = APP_DIR / "config.toml"


class Settings(BaseSettings):
    name: str = APP_NAME
    description: str = "Chercher, the universal and personal search engine."
    version: str = version(APP_NAME)

    db_url: str = f"{APP_DIR}/db.sqlite3"

    model_config = SettingsConfigDict(
        toml_file=CONFIG_FILE_PATH,
        extra="allow",
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        APP_DIR.mkdir(parents=True, exist_ok=True)

        return (
            init_settings,
            TomlConfigSettingsSource(settings_cls),
            env_settings,
            dotenv_settings,
            file_secret_settings,
        )


if __name__ == "__main__":
    settings = Settings()
    print(settings.model_dump_json())



import sqlite3
from contextlib import contextmanager


def init_db(conn: sqlite3.Connection) -> None:
    cursor = conn.cursor()

    # Create the documents table.
    cursor.execute("""
    CREATE TABLE IF NOT EXISTS documents (
        uri TEXT PRIMARY KEY,
        body TEXT NOT NULL,
        hash TEXT,
        metadata TEXT,

        created_at TEXT NOT NULL DEFAULT CURRENT_TIMESTAMP,
        last_updated_at TEXT NOT NULL DEFAULT CURRENT_TIMESTAMP
    );
    """)

    # Create virtual FTS5 table based on the documents table.
    cursor.execute("""
    CREATE VIRTUAL TABLE IF NOT EXISTS documents_fts USING fts5 (
        uri,
        body,
        content=documents,
        tokenize="porter unicode61 remove_diacritics 1"
    );
    """)

    # Create trigger to keep the FTS5 index up to date.
    cursor.execute("""
    CREATE TRIGGER IF NOT EXISTS documents_ai AFTER INSERT ON documents BEGIN
        INSERT INTO documents_fts(rowid, uri, body) VALUES (new.rowid, new.uri, new.body);
    END;
    """)

    # Settings
    cursor.execute("PRAGMA journal_mode = WAL;")
    cursor.execute("PRAGMA foreign_keys = ON;")
    cursor.execute("PRAGMA synchronous = NORMAL;")
    cursor.execute("PRAGMA busy_timeout = 5000;")
    cursor.execute("PRAGMA temp_store = FILE;")
    cursor.execute("PRAGMA cache_size = 2000;")
    cursor.execute("PRAGMA auto_vacuum = FULL;")

    conn.commit()


@contextmanager
def db_connection(db_url: str):
    conn = sqlite3.connect(db_url)

    try:
        yield conn
        conn.commit()
    except Exception as e:
        conn.rollback()
        raise e
    finally:
        conn.close()



from faker import Faker

fake = Faker()


def test_init_db(test_db):
    cursor = test_db.cursor()
    cursor.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='documents';"
    )
    assert cursor.fetchone() is not None

    cursor.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='documents_fts';"
    )
    assert cursor.fetchone() is not None


def test_document_insertion(test_db):
    uri = fake.file_path(depth=3)
    cursor = test_db.cursor()
    cursor.execute(
        "INSERT INTO documents (uri, body, metadata) VALUES (?, ?, ?)",
        (uri, "", "{}"),
    )
    test_db.commit()

    cursor.execute("SELECT * FROM documents WHERE uri = ?", (uri,))
    document = cursor.fetchone()
    assert document[0] == uri

    cursor.execute("SELECT * FROM documents_fts WHERE uri = ?", (uri,))
    document = cursor.fetchone()
    assert document[0] == uri



from pluggy import PluginManager
from chercher import hookspecs
from chercher.settings import APP_NAME


def get_plugin_manager() -> PluginManager:
    pm = PluginManager(APP_NAME)
    pm.add_hookspecs(hookspecs)
    pm.load_setuptools_entrypoints(APP_NAME)

    return pm
